from qbz95.ml.planar_utils import load_planar_dataset
from qbz95.ml.model import Model
from qbz95.ml.utils import *



class NeuronLayer(object):
    def __init__(self, num_neuron, num_weights):
        self.num_neuron = num_neuron
        self.num_weights = num_weights

    def forward_propagation(self, A):
        pass

    def backward_propagation(self, pre_layer, next_layer):
        pass


class InputLayer(NeuronLayer):
    def __init__(self, feature_num):
        NeuronLayer.__init__(self, feature_num, 0)
        self.A = None

    def forward_propagation(self, A, use_keep_prob=True):
        assert (A.shape[0] == self.num_neuron)
        self.A = A

    def print_layer(self):
        if self.A is not None:
            logger.info("A=" + str(self.A))


class HiddenLayer(NeuronLayer):

    def __init__(self, num_neuron, num_weights, activator=Sigmoid,
                 initializer=StaticWeightInitializer(0.01), regularizer=Regularizer(),
                 keep_prob=1, optimizer = GradientOptimizer(0.01)):
        assert(keep_prob <= 1 and keep_prob > 0)
        NeuronLayer.__init__(self, num_neuron, num_weights)
        self.activator = activator
        self.W, self.b = initializer.initial_weights(num_neuron, num_weights)
        self.regularizer = regularizer
        self.keep_prob = keep_prob
        self.optimizer = optimizer
        self.A = None
        self.Z = None
        self.D = None
        self.dA = None
        self.dZ = None
        self.dW = None
        self.db = None
        self.t = 0
        self.has_initialize_optimizer = False

    def print_layer(self):
        logger.info("W=" + str(self.W))
        logger.info("b=" + str(self.b))
        if self.D is not None:
            logger.info("D=" + str(self.D))
        if self.Z is not None:
            logger.info("Z=" + str(self.Z))
        if self.A is not None:
            logger.info("A=" + str(self.A))

    def regularize(self, m):
        return self.regularizer.regularize(m, self.W)

    def forward_propagation(self, A, use_keep_prob=True):
        assert(A.shape[0] == self.num_weights)
        m = A.shape[1]
        self.Z = np.dot(self.W, A) + self.b
        self.A = self.activator.activate(self.Z)
        if use_keep_prob and self.keep_prob < 1:
            D = np.random.rand(self.A.shape[0], self.A.shape[1])
            # if self.D is not None:
            #     np.testing.assert_array_equal(np.int64(D < self.keep_prob), self.D)
            #     logger.info("D always is same after every epoch")
            self.D = np.int64(D < self.keep_prob)
            self.A = self.A * self.D/self.keep_prob

    def backward_propagation(self, pre_layer, next_layer):
        m = self.Z.shape[1]
        # self.dZ = self.activator.gradient_layer(self.Z, next_layer)
        self.dA = np.dot(next_layer.W.T, next_layer.dZ)
        self.dZ = np.multiply(self.dA, self.activator.gradient(self.Z))
        if self.keep_prob < 1 and self.D is not None:
            self.dZ = self.dZ * self.D/self.keep_prob
        self.dW = np.dot(self.dZ, pre_layer.A.T) + self.regularizer.gradient(m, self.W)
        self.db = np.sum(self.dZ, axis=1, keepdims=True)

    def update_parameters(self):
        if not self.has_initialize_optimizer:
            self.optimizer.initialize(self)
            self.has_initialize_optimizer = True
        self.optimizer.update_parameters(self)


class OutputLayer(HiddenLayer):

    def __init__(self, num_neuron, num_weights, activator=Sigmoid,
                 initializer=StaticWeightInitializer(0.01), regularizer=Regularizer(),
                 optimizer=GradientOptimizer(0.01)):
        HiddenLayer.__init__(self, num_neuron, num_weights, activator,
                    initializer=initializer, regularizer=regularizer,
                    keep_prob=1, optimizer=optimizer)
        # self.cost = None

    # def compute_cost(self, Y):
    #     self.cost = self.activator.compute_cost(self.A, Y)

    def backward_propagation(self, pre_layer, dZ):
        m = dZ.shape[1]
        self.dZ = dZ #self.activator.gradient_cost(self.A, Y)
        self.dW = np.dot(self.dZ, pre_layer.A.T) + self.regularizer.gradient(m, self.W)
        self.db = np.sum(self.dZ, axis=1, keepdims=True)


class NeuralNetwork(Model):

    def __init__(self,
                 layers, iterator=NumberIterator(100),
                 keep_cost_num_epoch=1000,
                 mini_batch=0,
                 cost_object=SigmoidCrossEntropy
                 ):
        Model.__init__(self)
        assert(len(layers) >= 2)
        assert(mini_batch >= 0)
        self.layers = layers
        self.num_layer = len(layers)
        self.input_layer = self.layers[0]
        self.output_layer = self.layers[self.num_layer-1]
        self.iterator = iterator
        self.keep_cost_num_epoch = keep_cost_num_epoch
        self.has_run_fit = False
        self.mini_batch = mini_batch
        self.costs=[]
        self.cost=None
        self.cost_object = cost_object
        self.debug_parameters("initialize")

    @classmethod
    def build_3layers(cls, X, num_hidden_neuron, Y, iterator=NumberIterator(100), learning_rate=0.001,
                      initializer=StaticWeightInitializer(0.01), regularizer=Regularizer(),
                      cost_object=SigmoidCrossEntropy):
        input = InputLayer(X.shape[0])
        hidden = HiddenLayer(num_hidden_neuron, input.num_neuron, activator=Tanh,
                             initializer=initializer, regularizer=regularizer,
                             optimizer=GradientOptimizer(learning_rate)
                             )
        output = OutputLayer(Y.shape[0], hidden.num_neuron, activator=Sigmoid,
                             initializer=initializer, regularizer=regularizer,
                             optimizer=GradientOptimizer(learning_rate))
        layers = [input, hidden, output]
        model = NeuralNetwork(layers, iterator=iterator, cost_object=cost_object)
        return model

    @classmethod
    def build(cls, num_neurons, activators=[Sigmoid], iterator=NumberIterator(100),
                      optimizer=GradientOptimizer(0.001), initializer=StaticWeightInitializer(0.01),
                      regularizer=Regularizer(), keep_prob=1, mini_batch=0,
                      keep_cost_num_epoch=1000, initialize_seed = 3, cost_object=SigmoidCrossEntropy):
        layers = []
        len_layer = len(num_neurons)
        len_activator = len(activators)
        assert (len_activator == 1 or len_activator == len_layer - 1)
        if initialize_seed > 0: np.random.seed(initialize_seed)
        for i in range(len_layer):

            if i == 0:
                layers.append(InputLayer(num_neurons[i]))
            else:
                if len(activators) == 1:
                    activator = activators[0]
                else:
                    activator = activators[i - 1]
                if i == len_layer - 1:
                    layers.append(OutputLayer(num_neurons[i], layers[i - 1].num_neuron, activator=activator,
                                              initializer=initializer, regularizer=regularizer,
                                              optimizer=optimizer))
                else:
                    layers.append(HiddenLayer(num_neurons[i], layers[i - 1].num_neuron, activator=activator,
                                              initializer=initializer, regularizer=regularizer,
                                              keep_prob=keep_prob, optimizer=optimizer))

        model = NeuralNetwork(layers, iterator=iterator, keep_cost_num_epoch=keep_cost_num_epoch,
                              mini_batch=mini_batch, cost_object=cost_object)
        return model

    @classmethod
    def build_layers(cls, num_neurons, activators=[Sigmoid], iterator=NumberIterator(100), learning_rate=0.001,
                     initializer=StaticWeightInitializer(0.01), regularizer=Regularizer(), keep_prob = 1, mini_batch=0,
                     keep_cost_num_epoch=1000):
        layers = []
        len_layer = len(num_neurons)
        len_activator = len(activators)
        assert(len_activator==1 or len_activator==len_layer-1)
        for i in range(len_layer):

            if i == 0:
                layers.append(InputLayer(num_neurons[i]))
            else:
                if len(activators) == 1:
                    activator = activators[0]
                else:
                    activator = activators[i-1]
                if i == len_layer-1:
                    layers.append(OutputLayer(num_neurons[i], layers[i-1].num_neuron, activator=activator,
                                              initializer=initializer, regularizer=regularizer,
                                              optimizer=GradientOptimizer(learning_rate)))
                else:
                    layers.append(HiddenLayer(num_neurons[i], layers[i-1].num_neuron, activator=activator,
                                              initializer=initializer, regularizer=regularizer,
                                              keep_prob=keep_prob, optimizer=GradientOptimizer(learning_rate)))

        model = NeuralNetwork(layers, iterator=iterator, keep_cost_num_epoch=keep_cost_num_epoch, mini_batch=mini_batch)
        return model

    def debug_double_line(self, title=""):
        if self.iterator.epoch_times < 1:
            logger.debug("======================{}======================".format(title))

    def debug_line(self, title=""):
        if self.iterator.epoch_times < 1:
            logger.debug("----------------------{}----------------------".format(title))

    def info_line(self, title=""):
        logger.info("----------------------{}----------------------".format(title))

    def debug_parameters(self, title):
        if logger.level == logging.DEBUG and self.iterator.epoch_times < 1:
            for i in range(1, len(self.layers)):
                self.debug_line("layer {} - {} - parameters".format(i, title))
                logger.debug("W=" + str(self.layers[i].W))
                logger.debug("b=" + str(self.layers[i].b))

    def debug_gradient(self, title):
        if logger.level == logging.DEBUG and self.iterator.epoch_times < 1:
            for i in range(1, len(self.layers)):
                self.debug_line("layer {} - {} - gradient".format(i, title))
                if self.layers[i].dA is not None:
                    logger.debug("dA[:, 0]=" + str(self.layers[i].dA[:, 0]))
                logger.debug("dZ[:, 0]=" + str(self.layers[i].dZ[:, 0]))
                logger.debug("dW=" + str(self.layers[i].dW))
                logger.debug("db=" + str(self.layers[i].db))

    def debug_output(self, title=""):
        if logger.level == logging.DEBUG and self.iterator.epoch_times < 1:
            for i in range(1, len(self.layers)):
                self.debug_line("layer {} - {} - output".format(i, title))
                logger.debug("Z[:, 0]=" + str(self.layers[i].Z[:, 0]))
                logger.debug("A[:, 0]=" + str(self.layers[i].A[:, 0]))

    def forward_propagation(self, X, use_keep_prob=True, keep_prob_seed=1):
        if use_keep_prob and keep_prob_seed > 0:
            np.random.seed(keep_prob_seed)   # mainly used for dropout matrix.
        for i in range(self.num_layer):
            if i == 0:
                self.layers[i].forward_propagation(X, use_keep_prob)
            else:
                self.layers[i].forward_propagation(self.layers[i-1].A, use_keep_prob)

    def backward_propagation(self, Y):
        i = self.num_layer - 2
        dZ = self.cost_object.gradient_cost(self.output_layer.A, Y)
        self.output_layer.backward_propagation(self.layers[i], dZ)
        while i > 0:
            pre_layer = self.layers[i - 1]
            next_layer = self.layers[i + 1]
            self.layers[i].backward_propagation(pre_layer, next_layer)
            i = i - 1

    def update_parameters(self):
        for i in range(1, self.num_layer):
            self.layers[i].update_parameters()

    def compute_cost(self, Y):
        regularize_cost = 0
        m = Y.shape[1]
        for i in range(1, self.num_layer):
            regularize_cost = regularize_cost + self.layers[i].regularize(m)
        self.cost = self.cost_object.cost(self.output_layer.A, Y) + regularize_cost

    def get_cost(self, X, Y):
        self.activate(X)
        return self.cost_object.cost(self.output_layer.A, Y)

    def get_batch(self, m):
        batch = m if self.mini_batch == 0 or self.mini_batch > m else self.mini_batch
        return batch

    def shuffle_data(self, batch, X, Y, shuffle_seed):
        if shuffle_seed > 0:
            np.random.seed(shuffle_seed)
        m = X.shape[1]
        if batch < m:
            permutation = list(np.random.permutation(m))
            shuffled_X = X[:, permutation]
            shuffled_Y = Y[:, permutation].reshape((1, m))
        else:
            shuffled_X = X
            shuffled_Y = Y
        return shuffled_X, shuffled_Y

    def fit(self, X, Y, print_cost=False, print_num_epoch=1000, shuffle_seed=11):
        self.has_run_fit = True
        self.costs = []

        m = X.shape[1]
        batch = self.get_batch(m)

        while self.iterator.iterate():
            shuffled_X, shuffled_Y = self.shuffle_data(batch, X, Y, shuffle_seed)
            shuffle_seed = shuffle_seed + 1
            start = 0
            while start < m:
                end = start + batch
                end = end if end<=m else m
                self.debug_double_line("start={}, end={}, m={}".format(start, end, m))
                X_last = shuffled_X[:, start:end]
                Y_last = shuffled_Y[:, start:end]
                self.fit_batch(X_last, Y_last)
                start = end
            if self.iterator.epoch_times % self.keep_cost_num_epoch == 0:
                self.costs.append(self.cost)
            if print_cost and self.iterator.epoch_times % print_num_epoch == 0:
                logger.info("Cost after epoch {}: {:.10f}".format(self.iterator.epoch_times, self.cost))

        self.forward_propagation(X_last)
        self.compute_cost(Y_last)
        if self.iterator.epoch_times % self.keep_cost_num_epoch == 0:
            self.costs.append(self.cost)
        if print_cost:
            logger.info("Final Cost after epoch {}: {:.10f}".format(self.iterator.epoch_times, self.cost))

    def fit_batch(self, X, Y):
        self.forward_propagation(X)
        self.debug_output("forward propagation")
        self.compute_cost(Y)
        self.backward_propagation(Y)
        self.debug_gradient("back propagation")
        self.update_parameters()
        self.debug_parameters("update_parameters")

        self.debug_line("cost={}".format(self.cost))

    def print_layers(self):
        for i in range(1, self.num_layer):
            NeuralNetwork.print_layer_line(i)
            self.layers[i].print_layer()

    def activate(self, X):
        """
        predict. ignore dropout when predict.
        :param X:
        :return:
        """
        assert (self.has_run_fit)
        self.forward_propagation(X, use_keep_prob=False)
        return self.output_layer.A


    def plot_costs(self, title=None, xlabel=None, ylabel="cost", costs=None):
        if xlabel is None:
            xlabel = 'epochs (per ' + str(self.keep_cost_num_epoch) +')'
        if title is None:
            title = "Learning rate =" + str(self.output_layer.optimizer.learning_rate)
        super().plot_costs(title, xlabel=xlabel, ylabel=ylabel, costs=costs)

    def gradient_check(self, X, Y, epsilon=1e-7):
        grad = []
        grad_approx = []
        self.forward_propagation(X)
        self.backward_propagation(Y)
        for i in range(1, self.num_layer):
            W = self.layers[i].W
            b = self.layers[i].b
            for j in range(self.layers[i].num_neuron):
                for k in range(self.layers[i].num_weights):
                    self.layers[i].W = np.copy(W)
                    self.layers[i].W[j, k] = self.layers[i].W[j, k] + epsilon
                    self.forward_propagation(X)
                    self.compute_cost(Y)
                    cost_plus = self.cost

                    self.layers[i].W = np.copy(W)
                    self.layers[i].W[j, k] = self.layers[i].W[j, k] - epsilon
                    self.forward_propagation(X)
                    self.compute_cost(Y)
                    cost_minus = self.cost

                    grad_approx.append((cost_plus - cost_minus) / 2 / epsilon)
                    grad.append(self.layers[i].dW[j, k])

                self.layers[i].W = W

                self.layers[i].b = np.copy(b)
                self.layers[i].b[j, 0] = self.layers[i].b[j, 0] + epsilon
                self.forward_propagation(X)
                self.compute_cost(Y)
                cost_plus = self.cost

                self.layers[i].b = np.copy(b)
                self.layers[i].b[j, 0] = self.layers[i].b[j, 0] - epsilon
                self.forward_propagation(X)
                self.compute_cost(Y)
                cost_minus = self.cost

                grad_approx.append((cost_plus - cost_minus) / 2 / epsilon)
                grad.append(self.layers[i].db[j, 0])

            self.layers[i].b = b
        grad_approx = np.asarray(grad_approx)
        grad = np.asarray(grad)
        numerator = np.sum(np.power(grad_approx - grad, 2))
        denominator = np.sum(np.power(grad_approx, 2)) + np.sum(np.power(grad, 2))
        difference = numerator / denominator

        if difference > 2e-7:
            logger.info("There is a mistake in the backward propagation! difference = " + str(difference) )
        else:
            logger.info("Your backward propagation works perfectly fine! difference = " + str(difference) )
        assert(difference <= 2e-7)
        return difference


if __name__ == "__main__":
    from qbz95.tests.ml.testCases_v2 import *
    X, Y = load_planar_dataset()
    logger.setLevel(logging.DEBUG)

    logger.info("logger.level={}".format(logger.level))
    logger.info(X.shape)
    logger.info(Y.shape)

    input_layer = InputLayer(X.shape[0])
    hidden_layer = HiddenLayer(5, input_layer.num_neuron, activator=Tanh)
    output_layer = OutputLayer(1, hidden_layer.num_neuron, activator=Sigmoid)
    layers = [input_layer, hidden_layer, output_layer]
    model = NeuralNetwork(layers, iterator=NumberIterator(100))
    model.debug_parameters("abc")
    X_assess, Y_assess, parameters = forward_propagation_test_case()
    logger.info((X_assess, Y_assess, parameters ))


from chatterbot import ChatBot
from chatterbot.trainers import ChatterBotCorpusTrainer
from chatterbot.trainers import ListTrainer
import os
class SimpleChat():

    def __init__(self):
        self.chatbot = ChatBot('myBot',
                               storage_adapter="chatterbot.storage.MongoDatabaseAdapter",
                               logic_adapters=[
                                   {
                                       'import_path': 'chatterbot.logic.BestMatch'
                                   },
                                   {
                                       'import_path': 'chatterbot.logic.LowConfidenceAdapter',
                                       'threshold': 0.6,
                                       'default_response': '正在学习中'
                                   }
                               ],
                               input_adapter="chatterbot.input.VariableInputTypeAdapter",
                               output_adapter="chatterbot.output.TerminalAdapter",
                               database_uri='mongodb://127.0.0.1:27017',
                               database='chatbot'
                               )

        self.chatbot.set_trainer(ListTrainer)
        self.chatbot.train("chatterbot.corpus.chinese")

    def get_response(self, info):
        return str(self.chatbot.get_response(info))

if __name__ == '__main__':
    print(os.path.dirname(os.path.abspath(__file__)))
    chat = SimpleChat()
    res = chat.get_response('你好')
    print(res)




from flask import Flask, render_template
from chatterbot import ChatBot
from chatterbot.trainers import ChatterBotCorpusTrainer

app = Flask(__name__)

english_bot = ChatBot("English Bot",
                      storage_adapter="chatterbot.storage.MongoDatabaseAdapter",
                      database='chatterbot-flask',
                      database_uri='mongodb://172.17.0.1:27017'
                      # ,                      read_only=True
                      )

english_bot.set_trainer(ChatterBotCorpusTrainer)
english_bot.train("chatterbot.corpus.english")


@app.route("/")
def home():
    return render_template("index.html")

@app.route("/get/<string:query>")
def get_raw_response(query):
    return str(english_bot.get_response(query))


if __name__ == "__main__":
    app.run("aa01", port=9000, debug=True)


import numpy as np
import time
from qbz95.ml2.estimator import *
from qbz95.ml2.layer import *
from qbz95.ml2.utils import *


class Network(object):

    def __init__(self, layers, estimator=SoftmaxCrossEntropy()):
        self.layers = layers
        self.estimator = estimator

    def forward(self, X, train_flag=True):
        """
        Compute activations of all network layers by applying them sequentially.
        Return a list of activations for each layer.
        Make sure last activation corresponds to network logits.
        """
        activations = []
        input = X

        for layer in self.layers:
            input = layer.forward(input, train_flag)
            activations.append(input)

        assert len(activations) == len(self.layers)
        output = self.estimator.forward(input)

        return activations, output

    def evaluate(self, X, y):
        _, output = self.forward(X, train_flag=False)
        predict_y, predict_label_y, loss, accuracy = self.estimator.evaluate(output, y)
        return predict_y, predict_label_y, loss, accuracy

    def predict(self, X):
        _, output = self.forward(X, train_flag=False)
        predict_y, predict_label_y = self.estimator.predict(output)
        return predict_y, predict_label_y

    def _train(self, X, Y):
        """
        Train your network on a given batch of X and y.
        You first need to run forward to get all layer activations.
        Then you can run layer.backward going from last to first layer.

        After you called backward for all layers, all Dense layers have already made one gradient step.
        """
        # very ingenious

        layer_activations, _ = self.forward(X)
        layer_inputs = [X] + layer_activations  
        layer_output = layer_activations[-1]

        loss_grad = self.estimator.backward(layer_output, Y)

        for i in reversed(range(len(self.layers))):
            loss_grad = self.layers[i].backward(layer_inputs[i], loss_grad)

    def train(self, X_train, Y_train, X_val=None, Y_val=None, epochs=10, batch_size=0, log_per_epoch=1):
        logs = {'train loss':[], 'train accuracy':[], 'val loss':[], 'val accuracy':[], 'over-fitting':[]}
        for epoch in range(epochs):
            start_time = time.time()
            for X_batch,Y_batch in iterate_minibatches(X_train, Y_train, batch_size=batch_size,shuffle=True):
                self._train(X_batch, Y_batch)

            _, _, train_loss, train_accuracy = self.evaluate(X_train, Y_train)
            _, _, val_loss, val_accuracy = self.evaluate(X_val, Y_val)

            log = "epoch: {} | {:.2f} seconds | " + \
                  "train loss: {:.4f} | train accuracy: {:.4f} | " + \
                  "val loss: {:.4f} | val accuracy: {:.4f} | " + \
                  "over-fitting: {:.4f}"

            print(log.format(epoch, time.time() - start_time, train_loss, train_accuracy,
                             val_loss, val_accuracy, train_accuracy - val_accuracy))

            if epoch % log_per_epoch == 0:
                logs['train loss'].append(train_loss)
                logs['train accuracy'].append(train_accuracy)
                logs['val loss'].append(val_loss)
                logs['val accuracy'].append(val_accuracy)
                logs['over-fitting'].append(train_accuracy - val_accuracy)

        return logs



import numpy as np


class Estimator:

    def __init__(self):
        pass

    def forward(self, input):
        pass

    def loss_output(self, output, y):
        return np.sum(-np.log(output[y == 1])) / y.shape[0]

    def loss(self, input, y):
        output = self.forward(input)
        return self.loss_output(output, y)

    def backward(self, input, y):
        pass

    def evaluate(self, input, y):
        output = self.forward(input)
        predict_label_y = np.argmax(output, axis=-1)
        predict_y = np.eye(output.shape[1])[predict_label_y]

        loss = self.loss_output(output, y)
        label_y = np.argmax(y, axis=-1)
        accuracy = np.mean(predict_label_y == label_y)
        return predict_y, predict_label_y, loss, accuracy

    def predict(self, input):
        output = self.forward(input)
        predict_label_y = np.argmax(output, axis=-1)
        predict_y = np.eye(output.shape[0])[predict_label_y]

        return predict_y, predict_label_y


class CrossEntropy(Estimator):
    """
    if input is very small, -y/(input*y.shape[0]) maybe makes the result unstable. 
    because of it, we create softmax cross entropy estimator.
    """

    def __init__(self):
        Estimator.__init__(self)

    def forward(self, input):
        return input

    def backward(self, input, y):
        return -y/(input*y.shape[0])


class SoftmaxCrossEntropy(Estimator):
    """
    create softmax cross entropy estimator in order to make the numbers stable.
    """

    def __init__(self):
        Estimator.__init__(self)
        self._output = None

    def forward(self, input):
        e = np.exp(input - np.max(input, axis=-1, keepdims=True))
        s = np.sum(e, axis=-1, keepdims=True)
        self._output = e / s
        return self._output

    def backward(self, input, y):
        if self._output is None: self.forward(input)
        return (self._output - y)/y.shape[0]




import insuranceqa_data as insuranceqa

train_data = insuranceqa.load_pool_train()
# test_data = insuranceqa.load_pool_test()
# valid_data = insuranceqa.load_pool_valid()
answers_data = insuranceqa.load_pool_answers()

lines = []
with open('insuranceqa.json', 'w') as f:
    for i in range(len(train_data)):
        # print('-------------------------------------------------')
        x = train_data[str(i)]
        question = x['zh']
        answers = x['answers']
        # print("question={}, answers={}".format(question, answers))
        answer = answers_data[answers[0]]['zh']
        line = '"question_id":{0},"question":"{1}","answer":"{2}"'.format(i, question, answer)
        lines.append('{"index":{}}' + '\n{' + line + '}\n')
    content = "".join(lines)
    f.write(content)

# for i in range(1):
#     print('-------------------------------------------------')
#     x = answers_data[str(26078)]
#     print(x)

# valid_data, test_data and train_data share the same properties
# for x in train_data:
#     print('index %s value: %s ++$++ %s ++$++ %s' % \
#           (x, d[x]['zh'], d[x]['en'], d[x]['answers'], d[x]['negatives']))


# print(answers_data)
# for x in answers_data:
#     print('index %s: %s ++$++ %s' % (x, d[x]['zh'], d[x]['en']))


import numpy as np


class Regularizer(object):

    def __init__(self, alpha=0):
        self.alpha = alpha

    def regularize(self, weights):
        return 0

    def gradient(self, weights):
        return 0


class L2Regularizer(Regularizer):

    def __init__(self, alpha=0.001):
        Regularizer.__init__(self, alpha)

    def regularize(self, weights):
        return self.alpha * np.sum(np.multiply(weights, weights)) / 2

    def gradient(self, weights):
        return self.alpha * weights


class L1Regularizer(Regularizer):

    def __init__(self, alpha=0.001):
        Regularizer.__init__(self, alpha)

    def regularize(self, weights):
        return self.alpha * np.sum(np.abs(weights))

    def gradient(self, weights):
        return self.alpha * weights / np.abs(weights)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from qbz95.ml.utils import logger
from qbz95.ml.planar_utils import plot_decision_boundary
from qbz95.ml.model import Model
from sklearn import metrics


class Predictor(object):
    def __init__(self, model=Model()):
        self.model = model

    def predict(self, X):
        logger.debug("X.shape={}".format(X.shape))
        A = self.model.activate(X)
        logger.debug("A.shape={}".format(A.shape))
        return self.predict_(A)

    def predict_(self, A):
        return A

    def plot_regression_curve(self, X, y, title="regression", X_model=None):
        '''
        only work for X which only have 2 features.
        :param X:  (2, num_sample)
        :param y:  (1, num_sample)
        :param title:
        :return:
        '''
        if len(X.shape) == 2 and X.shape[0] == 1 and len(y.shape) == 2 and y.shape[0] == 1:
            plt.title(title)
            axes = plt.gca()
            min = X.min(axis=1)
            max = X.max(axis=1)
            range = max - min
            min = min - 0.1 * range
            max = max + 0.1 * range
            axes.set_xlim((min[0], max[0]))
            min = y.min(axis=1)
            max = y.max(axis=1)
            range = max - min
            min = min - 0.1 * range
            max = max + 0.1 * range
            axes.set_ylim((min[0], max[0]))
            y_predict =  self.predict(X) if X_model is None else self.predict(X_model)

            plt.scatter(X.T, y.T, c='C0', s=10, cmap=plt.cm.Spectral)

            order = np.lexsort((y_predict, X))
            X, y_predict = X[0, order], y_predict[0, order]
            plt.plot(X.T, y_predict.T, 'C1-')
            plt.show()
        else:
            print("plot_regression_curve only works for X which only have 1 features.")


class ClassificationPredictor(Predictor):

    def __init__(self, model=Model()):
        Predictor.__init__(self, model)

    def predict_(self, A):
        if A.shape[0] == 1:
            return np.int32(A > 0.5)
        else:
            a_max = np.max(A, axis=0, keepdims=True)
            return  np.int32(A == a_max)

    def accuracy(self, X, Y):
        Y_predict = self.predict(X)
        return self.accuracy_(Y_predict, Y)

    def accuracy_(self, Y_predict, Y):
        return np.mean(Y_predict == Y)

    def evaluate(self, X, Y, index = None):
        Y_predict = self.predict(X)
        accuracy = self.accuracy_(Y_predict, Y)
        cost = self.model.get_cost(X, Y)
        k = Y.shape[0]
        metrics = []
        for i in range(k):
            tp = np.sum((Y_predict[i]==1) * (Y[i]==1))
            fn = np.sum((Y_predict[i]==0) * (Y[i]==1))
            fp = np.sum((Y_predict[i]==1) * (Y[i]==0))
            tn = np.sum((Y_predict[i]==0) * (Y[i]==0))

            recall = tp/(tp+fn) if tp>0 else 0
            precision = tp/(tp+fp) if tp>0 else 0
            specificity = tn/(tn+fp) if tn>0 else 0
            f1 = 2*recall*precision/(recall+precision) if recall+precision>0 else 0

            metrics.append([accuracy, cost, recall, precision, specificity, f1, tp, fn, fp, tn])

        if index is None:
            index = range(k)
        metrics = pd.DataFrame(metrics, index=index,
                               columns=['accuracy', 'cost', 'recall', 'precision', 'specificity', 'f1',  'tp', 'fn', 'fp', 'tn'])
        return metrics

    def print_metrics(self, train_X, train_Y, dev_X=None, dev_Y=None, test_X=None, test_Y=None):
        metrics = []
        if train_X is not None:
            metrics_ = self.evaluate(train_X, train_Y)
            metrics_.index=["train"]
            metrics.append(metrics_)
        if dev_X is not None:
            metrics_ = self.evaluate(dev_X, dev_Y)
            metrics_.index = ["dev"]
            metrics.append(metrics_)
        if test_X is not None:
            metrics_ = self.evaluate(test_X, test_Y)
            metrics_.index = ["test"]
            metrics.append(metrics_)
        print(pd.concat(metrics))

    def print_accuracy(self, X, Y, title="train"):
        print("{} accuracy: {}".format(title, self.accuracy(X, Y)))

    def print_accuracy_train_test(self, train_X, train_Y, dev_X=None, dev_Y=None, test_X=None, test_Y=None):
        if train_X is not None:
            self.print_accuracy(train_X, train_Y, "train")
        if dev_X is not None:
            self.print_accuracy(dev_X, dev_Y, "dev")
        if test_X is not None:
            self.print_accuracy(test_X, test_Y, "test")

    def plot_decision_boundary(self, X, y, title="decision boundary"):
        '''
        only work for X which only have 2 features.
        :param X:  (2, num_sample)
        :param y:  (1, num_sample)
        :param title:
        :return:
        '''
        if len(X.shape)==2 and X.shape[0]==2 and len(y.shape)==2 and y.shape[0]==1:
            plt.title(title)
            axes = plt.gca()
            min = X.min(axis=1)
            max = X.max(axis=1)
            range = max - min
            min = min - 0.1 * range
            max = max + 0.1 * range
            axes.set_xlim((min[0], max[0]))
            axes.set_ylim((min[1], max[1]))
            plot_decision_boundary(lambda x: self.predict(x.T), X, y)
        else:
            print("plot_decision_boundary only works for X which only have 2 features.")


class ProbabilityPredictor(ClassificationPredictor):

    @classmethod
    def build(cls, train_Y, model=Model()):
        k = train_Y.shape[0]
        m = train_Y.shape[1]
        positive_prob = np.array([np.sum(train_Y[i] == 1)/m for i in range(k)]).reshape((k, 1))
        if k==1:
            prior_prob = positive_prob
        else:
            prior_prob = positive_prob/np.sum(positive_prob)
        return ProbabilityPredictor(prior_prob, model)

    def __init__(self, prior_prob=0.5, model=Model()):
        ClassificationPredictor.__init__(self, model)
        logger.debug("prior_prob={}".format(prior_prob))
        self.prior_prob = prior_prob

    def predict_(self, A):
        if A.shape[0] == 1:
            return np.int32(A > self.prior_prob)
        else:
            A_adjust = A/self.prior_prob
            logger.debug("A_adjust[0:20, 0:20]={}".format(A_adjust[0:20, 0:20]))
            A_adjust_max = np.max(A_adjust, axis=0, keepdims=True)
            return  np.int32(A_adjust == A_adjust_max)


class EvaluateCurve(object):
    # def __init__(self, model=Model()):
    #     Predictor.__init__(self, model)

    @classmethod
    def generate_data(cls, X, Y, model=Model(), cutoffs=np.linspace(1, 0, num=21)):
        assert(Y.shape[0]==1)
        A = model.activate(X)

        cutoff_metrics = []
        for cutoff in cutoffs:
            predictor = ProbabilityPredictor(cutoff, model)
            metrics = predictor.evaluate(X, Y)
            metrics.insert(0, "cut_off", cutoff)
            cutoff_metrics.append(metrics)

        cutoff_metrics = pd.concat(cutoff_metrics)
        cutoff_metrics.index = np.arange(0, len(cutoffs))
        return cutoff_metrics

    @classmethod
    def auc(cls, fpr=None, tpr=None, metrics=None):
        if metrics is None:
            return np.trapz(tpr, fpr)
        else:
            return np.trapz(metrics['recall'], 1 - metrics["specificity"])

    @classmethod
    def plot_roc(cls, cutoffs):
        plt.title('ROC Curve')
        plt.xlim((0, 1))
        plt.ylim((0, 1))
        plt.ylabel('Sensitivity(TPR)')
        plt.xlabel('1-Specificity(FPR)')
        plt.xticks(np.linspace(0, 1, 11))
        plt.yticks(np.linspace(0, 1, 11))

        m = len(cutoffs)
        i = 0
        for label, metrics in cutoffs.items():
            plt.plot(1-metrics["specificity"], metrics["recall"], 'C'+str(i % 10)+".-")
            i = i + 1
        plt.legend(cutoffs.keys(), loc='lower right')
        plt.plot([0,1],[0,1], '--', color='k')
        plt.show()

    @classmethod
    def plot_pr(cls, cutoffs):
        plt.title('Recall Precision')
        plt.xlim((0, 1))
        plt.ylim((0, 1))
        plt.ylabel('Recall')
        plt.xlabel('Precision')
        plt.xticks(np.linspace(0, 1, 11))
        plt.yticks(np.linspace(0, 1, 11))

        m = len(cutoffs)
        i = 0
        for label, metrics in cutoffs.items():
            plt.plot(metrics["precision"], metrics["recall"], 'C'+str(i % 10)+".-")
            i = i + 1
        plt.legend(cutoffs.keys(), loc='lower left')

        plt.show()



import numpy as np
from qbz95.rl.bandit.multi_armed_bandit import MultiArmedBandit, get_bandits
from qbz95.rl.bandit.epsilon_greedy_bandit import EpsilonGreedyBandit
from qbz95.rl.bandit.upper_limit_bandit import UpperLimitBandit
from qbz95.rl.bandit.delay_epsilon_bandit import DelayEpsilonBandit
import matplotlib.pyplot as plt


class UCBBandit(MultiArmedBandit):
    # Upper Confidence Bound

    def __init__(self, bandits):
        MultiArmedBandit.__init__(self, bandits)

    def pull(self):
        j = np.argmax([self.ucb(b) for b in self.bandits])
        return j, self.bandits[j].pull_update()

    def ucb(self, bandit):
        if bandit.n == 0:
            return float('inf')
        return bandit.mean_return + np.sqrt(2*np.log(len(self.returns)) / bandit.n)


if __name__ == "__main__":
    mab_bandit1 = UpperLimitBandit(get_bandits(10)).pulls(100000)
    mab_bandit2 = EpsilonGreedyBandit(get_bandits(), 0.1).pulls(100000)
    mab_bandit3 = UCBBandit(get_bandits()).pulls(100000)
    mab_bandit4 = DelayEpsilonBandit(get_bandits()).pulls(100000)
    mab_bandit1.plot_return(xscale="log", show=False, label='Upper Limit(ul=10)')
    mab_bandit2.plot_return(xscale="log", show=False, label='Epsilon Greedy(eps=0.1)')
    mab_bandit3.plot_return(xscale="log", show=False, label='UCB')
    mab_bandit4.plot_return(xscale="log", show=False, label='Delay Epsilon')
    plt.legend()
    plt.show()





from sanic import Sanic
from sanic.response import text, redirect


app = Sanic(__name__)


@app.route('/')
async def index(request):
    # generate a URL for the endpoint `post_handler`
    url = app.url_for('post_handler', post_id=5)
    # the URL is `/posts/5`, redirect to it
    return redirect(url)


@app.route('/posts/<post_id>')
async def post_handler(request, post_id):
    return text('Post - {}'.format(post_id))


@app.route('/tag/<tag>')
async def tag_handler(request, tag):
    return text('Tag - {}'.format(tag))


@app.route('/number/<integer_arg:int>')
async def integer_handler(request, integer_arg):
    return text('Integer - {}'.format(integer_arg))


@app.route('/number/<number_arg:number>')
async def number_handler(request, number_arg):
    return text('Number - {}'.format(number_arg))


@app.route('/person/<name:[A-z]>')
async def person_handler(request, name):
    return text('Person - {}'.format(name))


@app.route('/folder/<folder_id:[A-z0-9]{0,4}>')
async def folder_handler(request, folder_id):
    return text('Folder - {}'.format(folder_id))


app.run(host="127.0.0.1", port=8000, debug=True)


import unittest
from qbz95.ml2.estimator import *
from qbz95.ml2.utils import *
import numpy as np


# logger.setLevel(level=logging.DEBUG)


class TestLayer(unittest.TestCase):
    def test_CrossEntropy(self):

        # Test the gradients w.r.t. params
        def compute_out_given(input, y):
            l = CrossEntropy()
            return l.loss(input, y)

        def compute_grad(input, y):
            l = CrossEntropy()
            grad = l.backward(input, y)
            return grad

        m, n = 10,6
        # when input is very small, it will make the -y/(input*y.shape[0]) unstable.
        # therefore, we add a number 0.001 and set decimal 3 instead of 6
        input = np.abs(np.random.randn(m, n))+0.001
        y = np.eye(n)[np.random.randint(n, size=m)]

        numeric_grad = eval_numerical_gradient(lambda input: compute_out_given(input, y), input)
        grad = compute_grad(input, y)

        print("numeric_grad[0,:]=", numeric_grad[:, :])
        print("grad[0,:]=", grad[:, :])

        np.testing.assert_array_almost_equal(numeric_grad, grad, decimal=3,
                                             err_msg="input gradient does not match numeric input gradient")

    def test_SoftmaxCrossEntropy(self):

        # Test the gradients w.r.t. params
        def compute_out_given(input, y):
            l = SoftmaxCrossEntropy()
            return l.loss(input, y)

        def compute_grad(input, y):
            l = SoftmaxCrossEntropy()
            grad = l.backward(input, y)
            return grad

        m, n = 10,6
        # now the grad is very stable.
        input = np.abs(np.random.randn(m, n))
        y = np.eye(n)[np.random.randint(n, size=m)]

        numeric_grad = eval_numerical_gradient(lambda input: compute_out_given(input, y), input)
        grad = compute_grad(input, y)

        print("numeric_grad[0,:]=", numeric_grad[:, :])
        print("grad[0,:]=", grad[:, :])

        np.testing.assert_array_almost_equal(numeric_grad, grad, decimal=6,
                                             err_msg="input gradient does not match numeric input gradient")


import numpy as np
from qbz95.rl.bandit.multi_armed_bandit import MultiArmedBandit, get_bandits
from qbz95.rl.bandit.epsilon_greedy_bandit import EpsilonGreedyBandit
import matplotlib.pyplot as plt


class DelayEpsilonBandit(MultiArmedBandit):
    def __init__(self, bandits):
        MultiArmedBandit.__init__(self, bandits)

    def pull(self):
        p = np.random.random()
        if p < 1/(len(self.returns) + 1):
            j = np.random.choice(len(self.bandits))
        else:
            j = np.argmax([b.mean_return for b in self.bandits])
        return j, self.bandits[j].pull_update()


if __name__ == "__main__":
    mab_bandit1 = DelayEpsilonBandit(get_bandits()).pulls(100000)
    mab_bandit2 = EpsilonGreedyBandit(get_bandits(), 0.1).pulls(100000)
    mab_bandit1.plot_return(xscale="log", show=False, label='Delay Epsilon')
    mab_bandit2.plot_return(xscale="log", show=False, label='Epsilon Greedy(eps=0.1)')
    plt.legend()
    plt.show()





import torch
import torchvision
import torchvision.transforms as transforms

#The output of torchvision datasets are PILImage images of range [0, 1]. We transform them to Tensors of normalized range [-1, 1].
# torchvision datasets???[0, 1]? ????normalizationn?[-1, 1]
transform = transforms.Compose(
    [transforms.ToTensor(),
     transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

trainset = torchvision.datasets.CIFAR10(root='./data', train=True,
                                        download=True, transform=transform)


import numpy as np
from qbz95.rl.bandit.multi_armed_bandit import MultiArmedBandit, get_bandits
from qbz95.rl.bandit.epsilon_greedy_bandit import EpsilonGreedyBandit
from qbz95.rl.bandit.bandit import Bandit
import matplotlib.pyplot as plt


class UpperLimitBandit(MultiArmedBandit):
    def __init__(self, bandits):
        MultiArmedBandit.__init__(self, bandits)

    def pull(self):
        j = np.argmax([b.mean_return for b in self.bandits])
        return j, self.bandits[j].pull_update()


if __name__ == "__main__":
    mab_bandit1 = UpperLimitBandit(get_bandits(10)).pulls(100000)
    mab_bandit2 = EpsilonGreedyBandit(get_bandits(), 0.1).pulls(100000)
    mab_bandit1.plot_return(xscale="log", show=False, label='Upper Limit')
    mab_bandit2.plot_return(xscale="log", show=False, label='Epsilon Greedy(eps=0.1)')
    plt.legend()
    plt.show()





import unittest
from qbz95.ml2.layer import *
from qbz95.ml2.utils import *
import numpy as np


# logger.setLevel(level=logging.DEBUG)


class TestLayer(unittest.TestCase):
    def test_ReLU(self):
        np.random.seed(320)

        grad_output = np.random.randn(10, 32)
        x = np.random.randn(10, 32)
        l = ReLU()
        grads = l.backward(x, grad_output)
        numeric_grads = eval_numerical_gradient(lambda x: np.sum(l.forward(x) * grad_output), x=x)
        # print(grads[2,:])
        # print(numeric_grads[2,:])
        np.testing.assert_array_almost_equal(grads, numeric_grads, decimal=6, \
                                             err_msg = "gradient returned by your layer does not match the numerically computed gradient")

    def test_Sigmoid(self):
        np.random.seed(320)
        grad_output = np.random.randn(10, 32)
        x = np.linspace(-1, 1, 10 * 32).reshape([10, 32])
        l = Sigmoid()
        grads = l.backward(x, grad_output)
        numeric_grads = eval_numerical_gradient(lambda x: np.sum(l.forward(x) * grad_output), x=x)
        np.testing.assert_array_almost_equal(grads, numeric_grads, decimal=6, \
                                             err_msg="gradient returned by your layer does not match the numerically computed gradient")


    def test_Tanh(self):
        np.random.seed(320)
        grad_output = np.random.randn(10, 32)
        x = np.linspace(-1, 1, 10 * 32).reshape([10, 32])
        l = Tanh()
        grads = l.backward(x, grad_output)
        numeric_grads = eval_numerical_gradient(lambda x: np.sum(l.forward(x) * grad_output), x=x)
        np.testing.assert_array_almost_equal(grads, numeric_grads, decimal=6, \
                                             err_msg="gradient returned by your layer does not match the numerically computed gradient")

    def test_Dense(self):
        l = Dense(128, 150)

        assert -0.05 < l.weights.mean() < 0.05 and 1e-3 < l.weights.std() < 1e-1, \
            "The initial weights must have zero mean and small variance. " \
            "If you know what you're doing, remove this assertion."
        assert -0.05 < l.biases.mean() < 0.05, "Biases must be zero mean. Ignore if you have a reason to do otherwise."

        # To test the outputs, we explicitly set weights with fixed values. DO NOT DO THAT IN ACTUAL NETWORK!
        l = Dense(3, 4)

        x = np.linspace(-1, 1, 2 * 3).reshape([2, 3])
        l.weights = np.linspace(-1, 1, 3 * 4).reshape([3, 4])
        l.biases = np.linspace(-1, 1, 4)

        assert np.allclose(l.forward(x), np.array([[0.07272727, 0.41212121, 0.75151515, 1.09090909],
                                                   [-0.90909091, 0.08484848, 1.07878788, 2.07272727]]))
        print("Well done!")

        x = np.linspace(-1, 1, 10 * 32).reshape([10, 32])
        l = Dense(32, 64, optimizer=GradientOptimizer(learning_rate=0))

        numeric_grads = eval_numerical_gradient(lambda x: l.forward(x).sum(), x)
        grads = l.backward(x, np.ones([10, 64]))

        assert np.allclose(grads, numeric_grads, rtol=1e-3, atol=0), "input gradient does not match numeric grad"
        print("Well done!")

        # Test the gradients w.r.t. params
        def compute_out_given_wb(w, b):
            l = Dense(32, 64, optimizer=GradientOptimizer(learning_rate=1))
            l.weights = np.array(w)
            l.biases = np.array(b)
            x = np.linspace(-1, 1, 10 * 32).reshape([10, 32])
            return l.forward(x)

        def compute_grad_by_params(w, b):
            l = Dense(32, 64, optimizer=GradientOptimizer(learning_rate=1))
            l.weights = np.array(w)
            l.biases = np.array(b)
            x = np.linspace(-1, 1, 10 * 32).reshape([10, 32])
            l.backward(x, np.ones([10, 64]))
            return w - l.weights, b - l.biases

        w, b = np.random.randn(32, 64), np.linspace(-1, 1, 64)

        numeric_dw = eval_numerical_gradient(lambda w: compute_out_given_wb(w, b).sum(), w)
        numeric_db = eval_numerical_gradient(lambda b: compute_out_given_wb(w, b).sum(), b)
        grad_w, grad_b = compute_grad_by_params(w, b)

        np.testing.assert_array_almost_equal(numeric_dw, grad_w, decimal=6, err_msg="weight gradient does not match numeric weight gradient")
        np.testing.assert_array_almost_equal(numeric_db, grad_b, decimal=6, err_msg="weight gradient does not match numeric weight gradient")
        print("Well done!")

    def test_BatchNormalization(self):
        # Test the gradients w.r.t. params
        def compute_out_given(gamma, beta, x, grad_out):
            l = BatchNormalization()
            l.initial_params(gamma, beta)
            return np.sum(l.forward(x) * grad_out)

        def compute_grad(gamma, beta, x, grad_out):
            l = BatchNormalization()
            l.initial_params(gamma, beta)
            grad_x = l.backward(x, grad_out)
            return gamma - l.gamma, beta - l.beta, grad_x

        m, n = 16, 24
        gamma = np.random.randn(1, n)
        beta = np.random.randn(1, n)
        x = np.random.randn(m, n)
        grad_out = np.linspace(-1, 1, m * n).reshape(m, n)

        numeric_dgamma = eval_numerical_gradient(lambda gamma: compute_out_given(gamma, beta, x, grad_out), gamma)
        numeric_dbeta = eval_numerical_gradient(lambda beta: compute_out_given(gamma, beta, x, grad_out), beta)
        numeric_dx = eval_numerical_gradient(lambda x: compute_out_given(gamma, beta, x, grad_out), x)
        grad_gamma, grad_beta, grad_x = compute_grad(gamma, beta, x, grad_out)

        print("grad_x[0,:]=", grad_x[0, :])

        assert np.allclose(numeric_dgamma, grad_gamma, rtol=1e-3, \
                           atol=0), "gamma gradient does not match numeric gamma gradient"
        assert np.allclose(numeric_dbeta, grad_beta, rtol=1e-3, \
                           atol=0), "beta gradient does not match numeric beta gradient"
        assert np.allclose(numeric_dx, grad_x, rtol=1e-3, \
                           atol=0), "input gradient does not match numeric input gradient"

    def test_BatchNormalization1(self):
        """test mu and var of inference"""

        m, n, k = 16, 8, 10

        l = BatchNormalization()
        mu_array= np.zeros((k, n))
        var_array = np.zeros((k, n))
        mu_infer_array = np.zeros((k, n))
        var_infer_array = np.zeros((k, n))

        for i in range(k):
            input = np.random.permutation(m*n).reshape(m, n)

            mu, var, x = l.compute(input)
            l.forward(input)

            mu_array[i] = mu
            var_array[i] = var
            mu_infer_array[i] = l.mu
            var_infer_array[i] = l.var

        np.testing.assert_array_almost_equal(mu_infer_array, np.cumsum(mu_array, axis=0)/(np.arange(0, k)+1).reshape(k, 1), decimal=6,
                                             err_msg="mu_infer should be the mean of mu")
        np.testing.assert_array_almost_equal(var_infer_array, m/(m-1)*np.cumsum(var_array, axis=0)/(np.arange(0, k)+1).reshape(k, 1), decimal=6,
                                             err_msg="var_infer should be the mean of var")

    def test_Dropout(self):
        seed = 257
        # Test the gradients w.r.t. params
        def compute_out_given(x, grad_out):
            l = Dropout(drop_probability=0.5)
            np.random.seed(seed)
            return np.sum(l.forward(x) * grad_out)

        def compute_grad(x, grad_out):
            l = Dropout(drop_probability=0.5)
            np.random.seed(seed)
            grad_x = l.backward(x, grad_out)
            return grad_x

        m, n = 16, 24
        x = np.random.randn(m, n)
        grad_out = np.linspace(-1, 1, m * n).reshape(m, n)

        numeric_dx = eval_numerical_gradient(lambda x: compute_out_given(x, grad_out), x)
        grad_x = compute_grad(x, grad_out)

        print("grad_x[0,:]=", grad_x[0, :])

        np.testing.assert_array_almost_equal(numeric_dx, grad_x, decimal=6, err_msg="input gradient does not match numeric input gradient")


    def test_Softmax(self):
        seed = 257
        # Test the gradients w.r.t. params
        def compute_out_given(x, grad_out):
            l = Softmax()
            return np.sum(l.forward(x) * grad_out)

        def compute_grad(x, grad_out):
            l = Softmax()
            grad_x = l.backward(x, grad_out)
            return grad_x

        m, n = 16, 24
        x = np.random.randn(m, n)
        grad_out = np.linspace(-1, 1, m * n).reshape(m, n)

        numeric_dx = eval_numerical_gradient(lambda x: compute_out_given(x, grad_out), x)
        grad_x = compute_grad(x, grad_out)

        print("grad_x[0,:]=", grad_x[0, :])

        np.testing.assert_array_almost_equal(numeric_dx, grad_x, decimal=6, err_msg="input gradient does not match numeric input gradient")


import numpy as np
from qbz95.ml.utils import *
from qbz95.ml.model import Model


class LinearModel(Model):
    def __init__(self, num_features,
                 activator=Sigmoid, regularizer=Regularizer(), iterator=NumberIterator(100),
                 optimizer=GradientOptimizer(0.001), initializer=StaticWeightInitializer(0.01),
                 keep_cost_num_epoch=1000, cost_object=SigmoidCrossEntropy):
        """
            Args:
                num_features -- number of features
                activator -- activator
                regularizer -- regularizer
                iterator -- iterator
                optimizer -- optimizer
                initializer -- initializer
        """
        Model.__init__(self)
        self.num_features = num_features
        self.activator = activator
        self.regularizer = regularizer
        self.iterator = iterator
        self.optimizer = optimizer
        self.initializer = initializer
        self.mini_batch = 0
        self.keep_cost_num_epoch = keep_cost_num_epoch
        self.W, self.b = self.initializer.initial_weights(1, num_features)
        self.b = np.asscalar(self.b)
        self.cost = None
        self.dw = None
        self.db = None
        self.z = None
        self.a = None
        self.dz = None
        self.da = None
        self.has_run_fit = False
        self.t = 0
        self.cost_object = cost_object

    def compute_cost(self, X, y):
        m = X.shape[1]
        self.z = np.dot(self.W, X) + self.b
        self.a = self.activator.activate(self.z)
        self.cost = self.cost_object.cost(self.a, y) + self.regularizer.regularize(m, self.W)

    def activate(self, X):
        assert self.has_run_fit
        z = np.dot(self.W, X) + self.b
        return self.activator.activate(z)

    def get_cost(self, X, Y):
        return self.cost_object.cost(self.activate(X), Y)

    def propagate(self, X, y):
        m = X.shape[1]
        self.compute_cost(X, y)
        self.dz = self.cost_object.gradient_cost(self.a, y)
        self.dW = np.dot(self.dz, X.T) + self.regularizer.gradient(m, self.W)
        self.db = np.sum(self.dz, axis=1, keepdims=True)

        logger.debug("W={}, b={}".format(self.W, self.b))
        logger.debug("z[:,0]={}".format(self.z))
        logger.debug("a[:,0]={}".format(self.a))
        logger.debug("dW={}, db={}".format(self.dW, self.db))
        logger.debug("cost={}".format(self.cost))

    def fit(self, X, y, print_cost=False, print_num_epoch=1000):
        """
            Args:
                X -- data of shape (number of examples, number of features)
                y -- true "label" vector (containing 0 if non-cat, 1 if cat), of shape (number of examples, 1)
                print_cost --
        """
        assert(y.shape == (1, X.shape[1] ))
        assert(X.shape[0] == self.num_features)

        self.has_run_fit = True

        self.costs = []
        while self.iterator.iterate():
            self.propagate(X, y)
            self.update_parameters()

            if self.iterator.epoch_times % self.keep_cost_num_epoch == 0:
                self.costs.append(self.cost)
            if print_cost and self.iterator.epoch_times % print_num_epoch == 0:
                logger.info("Cost after epoch {}: {:.10f}".format(self.iterator.epoch_times, self.cost))
        self.compute_cost(X, y)
        if self.iterator.epoch_times % self.keep_cost_num_epoch == 0:
            self.costs.append(self.cost)
        if print_cost:
            logger.info("Final Cost after epoch {}: {:.10f}".format(self.iterator.epoch_times, self.cost))

    def update_parameters(self):
        self.optimizer.update_parameters(self)

    def plot_costs(self, title=None, xlabel=None, ylabel="cost", costs=None):
        if xlabel is None:
            xlabel = 'epochs (per ' + str(self.keep_cost_num_epoch) +')'
        if title is None:
            title = "Learning rate =" + str(self.optimizer.learning_rate)
        super().plot_costs(title, xlabel=xlabel, ylabel=ylabel, costs=costs)


class LogisticRegression(LinearModel):
    def __init__(self, num_features, activator=Sigmoid, regularizer=Regularizer(),
                 iterator=NumberIterator(100), optimizer=GradientOptimizer(0.001),
                 initializer=StaticWeightInitializer(0.01), keep_cost_num_epoch=1000,
                 cost_object=SigmoidCrossEntropy):
        LinearModel.__init__(self, num_features, activator=activator, regularizer=regularizer,
                             iterator=iterator, optimizer=optimizer,
                             initializer=initializer, keep_cost_num_epoch=keep_cost_num_epoch,
                             cost_object=cost_object)


class LinearRegression(LinearModel):
    def __init__(self, num_features, regularizer=Regularizer(),
                 iterator=NumberIterator(100), optimizer=GradientOptimizer(0.001),
                 initializer=WeightInitializer(), keep_cost_num_epoch=1000,
                 cost_object=MeanSquaredError
                 ):
        LinearModel.__init__(self, num_features, activator=Linear, regularizer=regularizer,
                             iterator=iterator, optimizer=optimizer,
                             initializer=initializer, keep_cost_num_epoch=keep_cost_num_epoch,
                             cost_object=cost_object)

    @classmethod
    def normal_equation(cls, X, y):
        X = add_intercept(X)
        theta = np.dot(y, np.dot(X.T, np.linalg.inv(np.dot(X, X.T)).T))
        b = theta[0:1, 0:1]
        w = theta[0:1, 1:]
        return w, b

if __name__ == "__main__":
    pass







import gym


env = gym.make('CartPole-v0')

print(env.action_space)
#> Discrete(2)
print(env.observation_space)
#> Box(4,)

print(env.observation_space.high)
#> array([ 2.4       ,         inf,  0.20943951,         inf])
print(env.observation_space.low)
#> array([-2.4       ,        -inf, -0.20943951,        -inf])
print(gym.envs.registry.all())
